==> cpufreq-LINUX_2_4-20040728.tar.gz: cpufreq/linux/arch/i386/kernel/cpufreq/speedstep-piix4.c <==
/*
 * (C) 2001  Dave Jones, Arjan van de ven.
 * (C) 2002 - 2003  Dominik Brodowski <linux@brodo.de>
 * (C) 2002 - 2003  Bruno Ducrot
 *
 *  Licensed under the terms of the GNU GPL License version 2.
 *
 *  BIG FAT DISCLAIMER: Work in progress code. Possibly *dangerous*
 */


/*********************************************************************
 *                        SPEEDSTEP - DEFINITIONS                    *
 *********************************************************************/

#include <linux/kernel.h>
#include <linux/module.h> 
#include <linux/init.h>
#include <linux/cpufreq.h>
#include <linux/pci.h>
#include <linux/slab.h>

#include "speedstep-lib.h"


/* speedstep_chipset:
 *   It is necessary to know which chipset is used. As accesses to 
 * this device occur at various places in this module, we need a 
 * static struct pci_dev * pointing to that device.
 */
static struct pci_dev			*speedstep_chipset_dev;


/* speedstep_processor
 */
static unsigned int			speedstep_processor = 0;


/* the GPO line the Intel SpeedStep Control Logic is connected to */
static int				gpo_hilo = -1;

/* 
 *   There are only two frequency states for each processor. Values
 * are in kHz for the time being.
 */
static struct cpufreq_frequency_table speedstep_freqs[] = {
	{SPEEDSTEP_HIGH, 	0},
	{SPEEDSTEP_LOW,		0},
	{0,			CPUFREQ_TABLE_END},
};


/* DEBUG
 *   Define it if you want verbose debug output, e.g. for bug reporting
 */
//#define SPEEDSTEP_DEBUG

#ifdef SPEEDSTEP_DEBUG
#define dprintk(msg...) printk(msg)
#else
#define dprintk(msg...) do { } while(0)
#endif


/**
 * speedstep_set_state - set the SpeedStep state
 * @state: new processor frequency state (SPEEDSTEP_LOW or SPEEDSTEP_HIGH)
 *
 *   Tries to change the SpeedStep state. 
 */
static void speedstep_set_state (unsigned int state, unsigned int notify)
{
	u32			pmbase;
	u32			pcntrl, pcntrl_save;
	u32			cntb, cntb_save;
	u32			val32;
	u8			pm2_blk;
	unsigned long		flags;
	struct cpufreq_freqs	freqs;

	if (!speedstep_chipset_dev || (state > 0x1))
		return;

	freqs.old = speedstep_get_processor_frequency(speedstep_processor);
	freqs.new = speedstep_freqs[state].frequency;
	freqs.cpu = 0; /* speedstep.c is UP only driver */
	
	if (notify)
		cpufreq_notify_transition(&freqs, CPUFREQ_PRECHANGE);

	/* get PMBASE */
	pci_read_config_dword(speedstep_chipset_dev, 0x40, &pmbase);
	if (!(pmbase & 0x01)) {
		printk(KERN_ERR "cpufreq: could not find speedstep register\n");
		return;
	}

	pmbase &= 0xFFFFFFFE;
	if (!pmbase) {
		printk(KERN_ERR "cpufreq: could not find speedstep register\n");
		return;
	}

	/* Disable IRQs */
	local_irq_save(flags);

	pci_read_config_dword(speedstep_chipset_dev, 0x48, &cntb);
	cntb_save = cntb;

	cntb &= ~(1 << 14);     /* ZZ_EN = 0 */
	cntb &= ~(0x0000001f << 6);
	cntb |= (0x1f << 6);    /* CPU_LCK */
	cntb |= (1 << 5);       /* CPU_SEL = 1 */

	dprintk(KERN_DEBUG "cntb before: 0x%.8x\n", cntb_save);
	dprintk(KERN_DEBUG "cntb after:  0x%.8x\n", cntb);
	pci_write_config_dword(speedstep_chipset_dev, 0x48, cntb);

	/* read state */
	val32 = inl_p(pmbase + 0x34);
	dprintk(KERN_DEBUG "reading %.8x from %.4x\n", val32, pmbase + 0x34);

	/* write new state */
	val32 &= ~(1 << gpo_hilo);
	val32 |= state << gpo_hilo;
	dprintk(KERN_DEBUG "new gpo: %.8x\n", val32);

	/* disable bus master arbitration */
	pm2_blk = inb(0x22);
	pm2_blk |= 0x01;
	outb_p(pm2_blk, 0x22);

	/* program southbridge for a deep sleep transition */
	pcntrl_save = pcntrl = inl(pmbase + 0x10);

	pcntrl &= ~(1 << 13);           /* CLKRUN_EN = 0 */
	pcntrl |= 1 << 12;              /* STPCLK_EN = 1 */
	pcntrl |= 1 << 11;              /* SLEEP_EN = 1 */
	pcntrl &= ~(1 << 10);           /* BURST_EN = 0 */
	pcntrl |= 1 << 9;               /* CC_EN = 1 */
	pcntrl &= ~(1 << 4);            /* THT_EN = 0 */

	dprintk(KERN_DEBUG "pcntrl before: 0x%.8x\n", pcntrl_save);
	dprintk(KERN_DEBUG "pcntrl after:  0x%.8x\n", pcntrl);

	outl_p(pcntrl, pmbase + 0x10);

	/* Write to GPO (HI/LO)#, then enter C3. */
	outl_p(val32, pmbase + 0x34);
	wbinvd();
	inb_p(pmbase + 0x15);

	/* We may have to do a dummy read after C3. */
	inb(0x80);

	/* cleanup */

	outl_p(pcntrl_save, pmbase + 0x10);
	/* restore bus master arbitration */

	pm2_blk &= 0xfe;
	outb_p(pm2_blk, 0x22);

	/* cleanup southbridge */
	pci_write_config_dword(speedstep_chipset_dev, 0x48, cntb_save);

	/* enable IRQs */

	local_irq_restore(flags);

	if (speedstep_get_processor_frequency(speedstep_processor) != freqs.old) {
		dprintk(KERN_INFO "cpufreq: change to %u MHz succeded\n", (freqs.new / 1000));
	} else {
		printk(KERN_ERR "cpufreq: change failed\n");
	}

	if (notify)
		cpufreq_notify_transition(&freqs, CPUFREQ_POSTCHANGE);

	return;
}


/**
 * speedstep_detect_chipset - detect the Southbridge which contains SpeedStep logic
 *
 *   Detects PIIX4. The pci_dev points to the PM module which contains all
 * power-management functions. Returns the SPEEDSTEP_CHIPSET_-number for
 * the detected chipset, or zero on failure.
 */
static unsigned int speedstep_detect_chipset (void)
{
	speedstep_chipset_dev = pci_find_subsys(PCI_VENDOR_ID_INTEL,
			      PCI_DEVICE_ID_INTEL_82371AB_3,
			      PCI_ANY_ID,
			      PCI_ANY_ID,
			      NULL);
	if (speedstep_chipset_dev)
		return 1; /* 440BX */

	speedstep_chipset_dev = pci_find_subsys(PCI_VENDOR_ID_INTEL,
			      PCI_DEVICE_ID_INTEL_82443MX_3,
			      PCI_ANY_ID,
			      PCI_ANY_ID,
			      NULL);
	if (speedstep_chipset_dev)
		return 1; /* 440MX */

	return 0;
}


/**
 * speedstep_setpolicy - set a new CPUFreq policy
 * @policy: new policy
 *
 * Sets a new CPUFreq policy.
 */
static int speedstep_target (struct cpufreq_policy *policy,
			     unsigned int target_freq,
			     unsigned int relation)
{
	unsigned int	newstate = 0;

	if (cpufreq_frequency_table_target(policy, &speedstep_freqs[0], target_freq, relation, &newstate))
		return -EINVAL;

	speedstep_set_state(newstate, 1);

	return 0;
}


/**
 * speedstep_verify - verifies a new CPUFreq policy
 * @freq: new policy
 *
 * Limit must be within speedstep_low_freq and speedstep_high_freq, with
 * at least one border included.
 */
static int speedstep_verify (struct cpufreq_policy *policy)
{
	return cpufreq_frequency_table_verify(policy, &speedstep_freqs[0]);
}


static int speedstep_cpu_init(struct cpufreq_policy *policy)
{
	int		result = 0;
	unsigned int	speed;

	/* capability check */
	if (policy->cpu != 0)
		return -ENODEV;

	/* detect low and high frequency */
	result = speedstep_get_freqs(speedstep_processor,
				     &speedstep_freqs[SPEEDSTEP_LOW].frequency,
				     &speedstep_freqs[SPEEDSTEP_HIGH].frequency,
				     &speedstep_set_state);
	if (result)
		return result;

	/* get current speed setting */
	speed = speedstep_get_processor_frequency(speedstep_processor);
	if (!speed)
		return -EIO;

	dprintk(KERN_INFO "cpufreq: currently at %s speed setting - %i MHz\n", 
		(speed == speedstep_freqs[SPEEDSTEP_LOW].frequency) ? "low" : "high",
		(speed / 1000));

	/* cpuinfo and default policy values */
	policy->policy = (speed == speedstep_freqs[SPEEDSTEP_LOW].frequency) ? 
		CPUFREQ_POLICY_POWERSAVE : CPUFREQ_POLICY_PERFORMANCE;
	policy->cpuinfo.transition_latency = CPUFREQ_ETERNAL;
	policy->cur = speed;

	return cpufreq_frequency_table_cpuinfo(policy, &speedstep_freqs[0]);
}


static struct cpufreq_driver speedstep_driver = {
	.name		= "speedstep",
	.verify 	= speedstep_verify,
	.target 	= speedstep_target,
	.init		= speedstep_cpu_init,
};


/**
 * speedstep_init - initializes the SpeedStep CPUFreq driver
 *
 *   Initializes the SpeedStep support. Returns -ENODEV on unsupported
 * devices, -EINVAL on problems during initiatization, and zero on
 * success.
 */
static int __init speedstep_init(void)
{
	/* detect processor */
	speedstep_processor = speedstep_detect_processor();
	if (!speedstep_processor)
		return -ENODEV;

	/* detect chipset */
	if (!speedstep_detect_chipset()) {
		printk(KERN_INFO "cpufreq: Intel(R) SpeedStep(TM) for this chipset not (yet) available.\n");
		return -ENODEV;
	}

	/* activate speedstep support */
	if (gpo_hilo < 0 || gpo_hilo > 32) {
		printk(KERN_INFO "SpeedStep on PIIX4 southbridge need a gpo_hilo option.\n");
		return -EINVAL;
	}

#ifdef CONFIG_APM
	printk(KERN_WARNING "SpeedStep and APM seems to be problematic.  Continuing anyway.\n");
#endif

	return cpufreq_register_driver(&speedstep_driver);
}


/**
 * speedstep_exit - unregisters SpeedStep support
 *
 *   Unregisters SpeedStep support.
 */
static void __exit speedstep_exit(void)
{
	cpufreq_unregister_driver(&speedstep_driver);
}

MODULE_PARM (gpo_hilo, "i");

MODULE_AUTHOR ("Bruno Ducrot");
MODULE_DESCRIPTION ("Speedstep driver for Intel mobile processors on chipsets with PIIX4 southbridges.");
MODULE_LICENSE ("GPL");

module_init(speedstep_init);
module_exit(speedstep_exit);

==> cpufreq-LINUX_2_5-20031003.tar.gz: cpufreq/linux/arch/i386/kernel/cpufreq/speedstep-piix4.c <==
/*
 * (C) 2001  Dave Jones, Arjan van de ven.
 * (C) 2002 - 2003  Dominik Brodowski <linux@brodo.de>
 * (C) 2002 - 2003  Bruno Ducrot
 *
 *  Licensed under the terms of the GNU GPL License version 2.
 *
 *  BIG FAT DISCLAIMER: Work in progress code. Possibly *dangerous*
 */


/*********************************************************************
 *                        SPEEDSTEP - DEFINITIONS                    *
 *********************************************************************/

#include <linux/kernel.h>
#include <linux/module.h> 
#include <linux/init.h>
#include <linux/cpufreq.h>
#include <linux/pci.h>
#include <linux/slab.h>

#include "speedstep-lib.h"


/* speedstep_chipset:
 *   It is necessary to know which chipset is used. As accesses to 
 * this device occur at various places in this module, we need a 
 * static struct pci_dev * pointing to that device.
 */
static struct pci_dev			*speedstep_chipset_dev;


/* speedstep_processor
 */
static unsigned int			speedstep_processor = 0;


/* the GPO line the Intel SpeedStep Control Logic is connected to */
static int				gpo_hilo = -1;

/* 
 *   There are only two frequency states for each processor. Values
 * are in kHz for the time being.
 */
static struct cpufreq_frequency_table speedstep_freqs[] = {
	{SPEEDSTEP_HIGH, 	0},
	{SPEEDSTEP_LOW,		0},
	{0,			CPUFREQ_TABLE_END},
};


/* DEBUG
 *   Define it if you want verbose debug output, e.g. for bug reporting
 */
//#define SPEEDSTEP_DEBUG

#ifdef SPEEDSTEP_DEBUG
#define dprintk(msg...) printk(msg)
#else
#define dprintk(msg...) do { } while(0)
#endif


/**
 * speedstep_set_state - set the SpeedStep state
 * @state: new processor frequency state (SPEEDSTEP_LOW or SPEEDSTEP_HIGH)
 *
 *   Tries to change the SpeedStep state. 
 */
static void speedstep_set_state (unsigned int state, unsigned int notify)
{
	u32			pmbase;
	u32			pcntrl, pcntrl_save;
	u32			cntb, cntb_save;
	u32			val32;
	u8			pm2_blk;
	unsigned long		flags;
	struct cpufreq_freqs	freqs;

	if (!speedstep_chipset_dev || (state > 0x1))
		return;

	freqs.old = speedstep_get_processor_frequency(speedstep_processor);
	freqs.new = speedstep_freqs[state].frequency;
	freqs.cpu = 0; /* speedstep.c is UP only driver */
	
	if (notify)
		cpufreq_notify_transition(&freqs, CPUFREQ_PRECHANGE);

	/* get PMBASE */
	pci_read_config_dword(speedstep_chipset_dev, 0x40, &pmbase);
	if (!(pmbase & 0x01)) {
		printk(KERN_ERR "cpufreq: could not find speedstep register\n");
		return;
	}

	pmbase &= 0xFFFFFFFE;
	if (!pmbase) {
		printk(KERN_ERR "cpufreq: could not find speedstep register\n");
		return;
	}

	/* Disable IRQs */
	local_irq_save(flags);

	pci_read_config_dword(speedstep_chipset_dev, 0x48, &cntb);
	cntb_save = cntb;

	cntb &= ~(1 << 14);     /* ZZ_EN = 0 */
	cntb &= ~(0x0000001f << 6);
	cntb |= (0x1f << 6);    /* CPU_LCK */
	cntb |= (1 << 5);       /* CPU_SEL = 1 */

	dprintk(KERN_DEBUG "cntb before: 0x%.8x\n", cntb_save);
	dprintk(KERN_DEBUG "cntb after:  0x%.8x\n", cntb);
	pci_write_config_dword(speedstep_chipset_dev, 0x48, cntb);

	/* read state */
	val32 = inl_p(pmbase + 0x34);
	dprintk(KERN_DEBUG "reading %.8x from %.4x\n", val32, pmbase + 0x34);

	/* write new state */
	val32 &= ~(1 << gpo_hilo);
	val32 |= state << gpo_hilo;
	dprintk(KERN_DEBUG "new gpo: %.8x\n", val32);

	/* disable bus master arbitration */
	pm2_blk = inb(0x22);
	pm2_blk |= 0x01;
	outb_p(pm2_blk, 0x22);

	/* program southbridge for a deep sleep transition */
	pcntrl_save = pcntrl = inl(pmbase + 0x10);

	pcntrl &= ~(1 << 13);           /* CLKRUN_EN = 0 */
	pcntrl |= 1 << 12;              /* STPCLK_EN = 1 */
	pcntrl |= 1 << 11;              /* SLEEP_EN = 1 */
	pcntrl &= ~(1 << 10);           /* BURST_EN = 0 */
	pcntrl |= 1 << 9;               /* CC_EN = 1 */
	pcntrl &= ~(1 << 4);            /* THT_EN = 0 */

	dprintk(KERN_DEBUG "pcntrl before: 0x%.8x\n", pcntrl_save);
	dprintk(KERN_DEBUG "pcntrl after:  0x%.8x\n", pcntrl);

	outl_p(pcntrl, pmbase + 0x10);

	/* Write to GPO (HI/LO)#, then enter C3. */
	outl_p(val32, pmbase + 0x34);
	wbinvd();
	inb_p(pmbase + 0x15);

	/* We may have to do a dummy read after C3. */
	inb(0x80);

	/* cleanup */

	outl_p(pcntrl_save, pmbase + 0x10);
	/* restore bus master arbitration */

	pm2_blk &= 0xfe;
	outb_p(pm2_blk, 0x22);

	/* cleanup southbridge */
	pci_write_config_dword(speedstep_chipset_dev, 0x48, cntb_save);

	/* enable IRQs */

	local_irq_restore(flags);

	if (speedstep_get_processor_frequency(speedstep_processor) != freqs.old) {
		dprintk(KERN_INFO "cpufreq: change to %u MHz succeded\n", (freqs.new / 1000));
	} else {
		printk(KERN_ERR "cpufreq: change failed\n");
	}

	if (notify)
		cpufreq_notify_transition(&freqs, CPUFREQ_POSTCHANGE);

	return;
}


/**
 * speedstep_detect_chipset - detect the Southbridge which contains SpeedStep logic
 *
 *   Detects PIIX4. The pci_dev points to the PM module which contains all
 * power-management functions. Returns the SPEEDSTEP_CHIPSET_-number for
 * the detected chipset, or zero on failure.
 */
static unsigned int speedstep_detect_chipset (void)
{
	speedstep_chipset_dev = pci_find_subsys(PCI_VENDOR_ID_INTEL,
			      PCI_DEVICE_ID_INTEL_82371AB_3,
			      PCI_ANY_ID,
			      PCI_ANY_ID,
			      NULL);
	if (speedstep_chipset_dev)
		return 1; /* 440BX */

	speedstep_chipset_dev = pci_find_subsys(PCI_VENDOR_ID_INTEL,
			      PCI_DEVICE_ID_INTEL_82443MX_3,
			      PCI_ANY_ID,
			      PCI_ANY_ID,
			      NULL);
	if (speedstep_chipset_dev)
		return 1; /* 440MX */

	return 0;
}


/**
 * speedstep_setpolicy - set a new CPUFreq policy
 * @policy: new policy
 *
 * Sets a new CPUFreq policy.
 */
static int speedstep_target (struct cpufreq_policy *policy,
			     unsigned int target_freq,
			     unsigned int relation)
{
	unsigned int	newstate = 0;

	if (cpufreq_frequency_table_target(policy, &speedstep_freqs[0], target_freq, relation, &newstate))
		return -EINVAL;

	speedstep_set_state(newstate, 1);

	return 0;
}


/**
 * speedstep_verify - verifies a new CPUFreq policy
 * @freq: new policy
 *
 * Limit must be within speedstep_low_freq and speedstep_high_freq, with
 * at least one border included.
 */
static int speedstep_verify (struct cpufreq_policy *policy)
{
	return cpufreq_frequency_table_verify(policy, &speedstep_freqs[0]);
}


static int speedstep_cpu_init(struct cpufreq_policy *policy)
{
	int		result = 0;
	unsigned int	speed;

	/* capability check */
	if (policy->cpu != 0)
		return -ENODEV;

	/* detect low and high frequency */
	result = speedstep_get_freqs(speedstep_processor,
				     &speedstep_freqs[SPEEDSTEP_LOW].frequency,
				     &speedstep_freqs[SPEEDSTEP_HIGH].frequency,
				     &speedstep_set_state);
	if (result)
		return result;

	/* get current speed setting */
	speed = speedstep_get_processor_frequency(speedstep_processor);
	if (!speed)
		return -EIO;

	dprintk(KERN_INFO "cpufreq: currently at %s speed setting - %i MHz\n", 
		(speed == speedstep_freqs[SPEEDSTEP_LOW].frequency) ? "low" : "high",
		(speed / 1000));

	/* cpuinfo and default policy values */
	policy->policy = (speed == speedstep_freqs[SPEEDSTEP_LOW].frequency) ? 
		CPUFREQ_POLICY_POWERSAVE : CPUFREQ_POLICY_PERFORMANCE;
	policy->cpuinfo.transition_latency = CPUFREQ_ETERNAL;
	policy->cur = speed;

	return cpufreq_frequency_table_cpuinfo(policy, &speedstep_freqs[0]);
}


static struct cpufreq_driver speedstep_driver = {
	.name		= "speedstep-piix4",
	.verify 	= speedstep_verify,
	.target 	= speedstep_target,
	.init		= speedstep_cpu_init,
	.owner		= THIS_MODULE,
};


/**
 * speedstep_init - initializes the SpeedStep CPUFreq driver
 *
 *   Initializes the SpeedStep support. Returns -ENODEV on unsupported
 * devices, -EINVAL on problems during initiatization, and zero on
 * success.
 */
static int __init speedstep_init(void)
{
	/* detect processor */
	speedstep_processor = speedstep_detect_processor();
	if (!speedstep_processor)
		return -ENODEV;

	/* detect chipset */
	if (!speedstep_detect_chipset()) {
		printk(KERN_INFO "cpufreq: Intel(R) SpeedStep(TM) for this chipset not (yet) available.\n");
		return -ENODEV;
	}

	/* activate speedstep support */
	if (gpo_hilo < 0 || gpo_hilo > 32) {
		printk(KERN_INFO "SpeedStep on PIIX4 southbridge need a gpo_hilo option.\n");
		return -EINVAL;
	}

#ifdef CONFIG_APM
	printk(KERN_WARNING "SpeedStep and APM seems to be problematic.  Continuing anyway.\n");
#endif

	return cpufreq_register_driver(&speedstep_driver);
}


/**
 * speedstep_exit - unregisters SpeedStep support
 *
 *   Unregisters SpeedStep support.
 */
static void __exit speedstep_exit(void)
{
	cpufreq_unregister_driver(&speedstep_driver);
}

MODULE_PARM (gpo_hilo, "i");

MODULE_AUTHOR ("Bruno Ducrot");
MODULE_DESCRIPTION ("Speedstep driver for Intel mobile processors on chipsets with PIIX4 southbridges.");
MODULE_LICENSE ("GPL");

module_init(speedstep_init);
module_exit(speedstep_exit);
